//
// shatbox (C) 2001 Matthew Conte (matt@conte.com)
//
//
// cart.cpp
//
// NES rom file handling
//
// $Id: $
//

#include <iostream>
#include <stdlib.h>
#include <string.h>
#include "cart.h"
#include "ppu.h"

#ifndef PATH_MAX
#define PATH_MAX 512
#endif // !PATH_MAX

#define  VERSUS_PAL_LENGTH    (64 * 3)



static char *changeExt(char *pFilename, const char *pExt)
{
   if (strrchr(pFilename, '.'))
      strcpy(strrchr(pFilename, '.'), pExt);
   else
      strcpy(pFilename, pExt);

   return pFilename;
}


cart::cart()
{
   memset(&mMapperInfo, 0, sizeof(mapperInfo));
   mFlags = 0;
   mpFilename = NULL;
}
  

cart::~cart()
{
   unload();
}


int cart::loadCartType()
{
   if (isUNIF(mpFilename))
      return loadUNIF();
   else if (isINES(mpFilename))
      return loadINES();

   return -1;
}


int cart::loadSRAM()
{
   if (mFlags & CART_BATTERY)
   {
      FILE *fp;
      char savename[PATH_MAX + 1];

      strncpy(savename, mpFilename, PATH_MAX);
      changeExt(savename, ".sav");

      fp = fopen(savename, "rb");
      if (NULL == fp)
         return -1;

      if (fread(mMapperInfo.mpSRAM, SRAM_BANK_SIZE, mMapperInfo.mNumSRAM, fp) < mMapperInfo.mNumSRAM)
      {
         fclose(fp);
         return -1;
      }
      else
      {
         fclose(fp);
      }
   }

   return 0;
}

void cart::saveSRAM()
{
   if (mFlags & CART_BATTERY)
   {
      FILE *fp;
      char savename[PATH_MAX + 1];

      strncpy(savename, mpFilename, PATH_MAX);
      changeExt(savename, ".sav");

      fp = fopen(savename, "wb");
      if (NULL != fp)
      {
         if (fwrite(mMapperInfo.mpSRAM, SRAM_BANK_SIZE, mMapperInfo.mNumSRAM, fp) < mMapperInfo.mNumSRAM)
         {
            fclose(fp);
            remove(savename);
         }
         else
         {
            fclose(fp);
         }
      }
   }
}

// versus ROMs have their own special PPUs, and
// hence, their own palettes.
int cart::loadPalette()
{
   FILE *fp = NULL;

   mMapperInfo.mpPalette = NULL;

   if (mFlags & CART_VERSUS)
   {
      char palname[PATH_MAX + 1];

      strncpy(palname, mpFilename, PATH_MAX);
      changeExt(palname, ".mpal");

      fp = fopen(palname, "rb");
      if (NULL != fp)
      {
         mMapperInfo.mpPalette = new uint8[VERSUS_PAL_LENGTH];
         if (NULL == mMapperInfo.mpPalette)
            goto _error;

         if (fread(mMapperInfo.mpPalette, VERSUS_PAL_LENGTH, 1, fp) < 1)
            goto _error;
      }
   }

   return 0;

_error:
   if (fp)
      fclose(fp);

   if (mMapperInfo.mpPalette)
   {
      delete [] mMapperInfo.mpPalette;
      mMapperInfo.mpPalette = NULL;
   }

   std::cerr << "vs. unisystem detected, but no palette file available\n";

   return -1;
}


int cart::allocateRAM()
{
   // allocate SRAM/VRAM if necessary
   if (mMapperInfo.mNumSRAM)
   {
      mMapperInfo.mpSRAM = new uint8[mMapperInfo.mNumSRAM * SRAM_BANK_SIZE];
      if (NULL == mMapperInfo.mpSRAM)
         return -1;

      // clear
      memset(mMapperInfo.mpSRAM, 0, mMapperInfo.mNumSRAM * SRAM_BANK_SIZE);
   }

   if (mMapperInfo.mNumVRAM)
   {
      mMapperInfo.mpVRAM = new uint8[mMapperInfo.mNumVRAM * VRAM_BANK_SIZE];
      if (NULL == mMapperInfo.mpVRAM)
         return -1;

      // clear
      memset(mMapperInfo.mpVRAM, 0, mMapperInfo.mNumVRAM * VRAM_BANK_SIZE);
   }
   
   return 0;
}


void cart::showInfo()
{
   char *pMirror = 0;
   switch (mMapperInfo.mMirror)
   {
   case MIRROR_HORIZ:
      pMirror = "horizontal";
      break;
   case MIRROR_VERT:
      pMirror = "vertical";
      break;
   case MIRROR_ONE_LOW:
      pMirror = "one screen - $2000";
      break;
   case MIRROR_ONE_HIGH:
      pMirror = "one screen - $2400";
      break;
   case MIRROR_FOUR:
      pMirror = "four screen";
      break;
   case MIRROR_MAPPER:
      pMirror = "mapper specifiable";
      break;
   }

   std::cout << "ROM information:" << std::endl;
   std::cout << "filename:       " << mpFilename  << std::endl;
   std::cout << "PRG ROM:        " << mMapperInfo.mNumPRG * 16 << "kB" << std::endl;
   std::cout << "CHR ROM:        " << mMapperInfo.mNumCHR * 8 << "kB" << std::endl;

   std::cout << "mirror type:    " << pMirror << std::endl;

   std::cout << "mapper number:  " << mMapperInfo.mMapperID << std::endl;
   std::cout << "battery:        " << ((mFlags & CART_BATTERY) ? "yes" : "no") << std::endl;
   std::cout << "trainer:        " << ((mFlags & CART_TRAINER) ? "yes" : "no") << std::endl;
   std::cout << "vs. unisystem:  " << ((mFlags & CART_VERSUS) ? "yes" : "no") << std::endl;
}


// TODO: load up intro if we fail loading ROM
int cart::load(const char *pFilename)
{
   if (NULL == pFilename)
      return -1;

   // copy the filename
   mpFilename = strdup(pFilename);
   if (NULL == mpFilename)
      return -1;

   if (loadCartType())
      return -1;

   if (allocateRAM())
      return -1;

   if (loadSRAM())
      return -1;

   if (loadPalette())
      return -1;

   // show some info
   showInfo();

   return 0;
}

void cart::unload()
{
   saveSRAM();

   delete [] mpFilename;
   mpFilename = NULL;

   delete [] mMapperInfo.mpPRG;
   mMapperInfo.mpPRG = NULL;

   delete [] mMapperInfo.mpCHR;
   mMapperInfo.mpCHR = NULL;

   delete [] mMapperInfo.mpSRAM;
   mMapperInfo.mpSRAM = NULL;

   delete [] mMapperInfo.mpVRAM;
   mMapperInfo.mpVRAM = NULL;

   delete [] mMapperInfo.mpPalette;
   mMapperInfo.mpPalette = NULL;
}


mapperInfo *cart::getMapperInfo()
{
   return &mMapperInfo;
}

