//
// shatbox (C) 2002 Matthew Conte (matt@conte.com)
//
//
// mmc.cpp
//
// supported MMC interfaces
//
// $Id: $
//

#include <iostream>
#include <stdlib.h> // for rand
#include "assert.h"
#include "nes.h"
#include "mmc.h"
#include "ppu.h"
#include "nes6502.h"
#include "mmc/map000.h"
#include "mmc/map001.h"
#include "mmc/map002.h"
#include "mmc/map003.h"
#include "mmc/map004.h"
#include "mmc/map007.h"
#include "mmc/map008.h"
#include "mmc/map011.h"



mmc::mmc(machine *pMachine, mapperInfo *pInfo)
: unit(pMachine), mpName(NULL)
{
   mInfo = *pInfo;
}

mmc::~mmc()
{
}

void mmc::getState(SnssMapperBlock */*pState*/)
{
}

void mmc::setState(SnssMapperBlock */*pState*/)
{
}

void mmc::reset(bool hard_reset)
{
   // TODO: find a cleaner way to do this
   if (mInfo.mNumSRAM)
   {
      nes6502_context cpu;
      nes6502_getcontext(&cpu);

      cpu.mem_page[6] = mInfo.mpSRAM;
      cpu.mem_page[7] = mInfo.mpSRAM + NES6502_BANKSIZE;

      nes6502_setcontext(&cpu);
   }

   // trash VRAM
   // TODO: verify if this is kosher
   if (hard_reset)
   {
      uint32 i;

      for (i = 0; i < mInfo.mNumVRAM * VRAM_BANK_SIZE; i++)
         mInfo.mpVRAM[i] = (uint8) rand();
   }

   // set default PRG config for "most" carts
   // (mapper code can override later)
   prgbank(16, 0x8000, 0);
   prgbank(16, 0xc000, CART_LASTBANK);

   // if VROM, bank it in, else bank in VRAM
   if (mInfo.mNumCHR)
   {
      chrbank(8, 0x0000, 0);
   }
   else if (mInfo.mNumVRAM)
   {
      // TODO: this is not accurate
      ((NES *) mpMachine)->getPPU()->bank(8, 0, mInfo.mpVRAM, false);
   }
   else
   {
      std::cerr << "cart has no CHR-ROM or VRAM, this could get interesting..." << std::endl;
   }

   switch (mInfo.mMirror)
   {
   case MIRROR_HORIZ:
      ((NES *) mpMachine)->getPPU()->mirror(0, 0, 1, 1);
      break;

   case MIRROR_VERT:
      ((NES *) mpMachine)->getPPU()->mirror(0, 1, 0, 1);
      break;
      
   case MIRROR_ONE_LOW:
      ((NES *) mpMachine)->getPPU()->mirror(0, 0, 0, 0);
      break;

   case MIRROR_ONE_HIGH:
      ((NES *) mpMachine)->getPPU()->mirror(1, 1, 1, 1);
      break;

   case MIRROR_FOUR:
      ((NES *) mpMachine)->getPPU()->mirror(0, 1, 2, 3);
      break;

   case MIRROR_MAPPER:
      break;
   }
}


void mmc::vblank()
{
}


void mmc::hblank(bool /*in_vblank*/)
{
}

/* operation functions */
#define  PRG_32      (mInfo.mNumPRG / 2)
#define  PRG_16      (mInfo.mNumPRG)
#define  PRG_8       (mInfo.mNumPRG * 2)
#define  CHR_8       (mInfo.mNumCHR)
#define  CHR_4       (mInfo.mNumCHR * 2)
#define  CHR_2       (mInfo.mNumCHR * 4)
#define  CHR_1       (mInfo.mNumCHR * 8)

#define  LAST_PRG_32 (PRG_32 - 1)
#define  LAST_PRG_16 (PRG_16 - 1)
#define  LAST_PRG_8  (PRG_8 - 1)
#define  LAST_CHR_8  (CHR_8 - 1)
#define  LAST_CHR_4  (CHR_4 - 1)
#define  LAST_CHR_2  (CHR_2 - 1)
#define  LAST_CHR_1  (CHR_1 - 1)

/* cpu bankswitching */
void mmc::prgbank(int size, uint32 address, int bank)
{
   nes6502_context cpu;
   int page, count, i;
   uint8 *base;

   nes6502_getcontext(&cpu);

   switch (size)
   {
   case 32:
      if (CART_LASTBANK == bank)
         bank = LAST_PRG_32;

      base = mInfo.mpPRG + ((bank % PRG_32) * 32768);
      break;

   case 16:
      if (CART_LASTBANK == bank)
         bank = LAST_PRG_16;

      base = mInfo.mpPRG + ((bank % PRG_16) * 16384);
      break;

   case 8:
      if (CART_LASTBANK == bank)
         bank = LAST_PRG_8;

      base = mInfo.mpPRG + ((bank % PRG_8) * 8192);
      break;

   default:
      std::cerr << "mmc::prgbank: invalid size " << size << "\n";
      exit(-1);
      break;
   }

   page = address / NES6502_BANKSIZE;
   count = size / 4;

   ASSERT(page + count <= NES6502_NUMBANKS);
   
   for (i = 0; i < count; i++)
      cpu.mem_page[page + i] = base + (i * NES6502_BANKSIZE);

   nes6502_setcontext(&cpu);
}


void mmc::chrbank(int size, uint32 address, int bank)
{
   // some carts are dumb.
   if (0 == mInfo.mNumCHR)
      return;

   switch (size)
   {
   case 8:
      if (CART_LASTBANK == bank)
         bank = LAST_CHR_8;

      ((NES *) mpMachine)->getPPU()->bank(8, address / PPU_BANKSIZE, mInfo.mpCHR + ((bank % CHR_8) * 8 * PPU_BANKSIZE) - address, true);
      break;

   case 4:
      if (CART_LASTBANK == bank)
         bank = LAST_CHR_4;

      ((NES *) mpMachine)->getPPU()->bank(4, address / PPU_BANKSIZE, mInfo.mpCHR + ((bank % CHR_4) * 4 * PPU_BANKSIZE) - address, true);
      break;

   case 2:
      if (CART_LASTBANK == bank)
         bank = LAST_CHR_2;

      ((NES *) mpMachine)->getPPU()->bank(2, address / PPU_BANKSIZE, mInfo.mpCHR + ((bank % CHR_2) * 2 * PPU_BANKSIZE) - address, true);
      break;

   case 1:
      if (CART_LASTBANK == bank)
         bank = LAST_CHR_1;

      ((NES *) mpMachine)->getPPU()->bank(1, address / PPU_BANKSIZE, mInfo.mpCHR + ((bank % CHR_1) * PPU_BANKSIZE) - address, true);
      break;
   }
}


// mmc factory
mmc *mmc_create(machine *pMachine, mapperInfo *pInfo)
{
   mmc *pMMC = NULL;

   switch (pInfo->mMapperID)
   {
   case 0:
      pMMC = new map000(pMachine, pInfo);
      break;
   case 1:
      pMMC = new map001(pMachine, pInfo);
      break;
   case 2:
      pMMC = new map002(pMachine, pInfo);
      break;
   case 3:
      pMMC = new map003(pMachine, pInfo);
      break;
   case 4:
      pMMC = new map004(pMachine, pInfo);
      break;
   case 7:
      pMMC = new map007(pMachine, pInfo);
      break;
   case 8:
      pMMC = new map008(pMachine, pInfo);
      break;
   case 11:
      pMMC = new map011(pMachine, pInfo);
      break;
   default:
      break;
   }

   return pMMC;
}
