/*
// shatbox (C) 2002 Matthew Conte (matt@conte.com)
**
**
** map001.cpp
**
** mapper 1 (MMC1) implementation
**
** $Id: $
*/

#include "mmc/map001.h"
#include "nes.h"
#include "ppu.h"

/* TODO: WRAM enable ala Mark Knibbs:
   ==================================
The SNROM board uses 8K CHR-RAM. The CHR-RAM is paged (i.e. it can be split
into two 4Kbyte pages).

The CRA16 line of the MMC1 is connected to the /CS1 pin of the WRAM. THIS MEANS
THAT THE WRAM CAN BE ENABLED OR DISABLED ACCORDING TO THE STATE OF THE CRA16
LINE. The CRA16 line corresponds to bit 4 of MMC1 registers 1 & 2.

The WRAM is enabled when the CRA16 line is low, and disabled when CRA16 is
high. This has implications when CHR page size is 4K, if the two page numbers
set have different CRA16 states (e.g. reg 1 bit 4 = 0, reg 2 bit 4 = 1). Then
the WRAM will be enabled and disabled depending on what CHR address is being
accessed.

When CHR page size is 8K, bit 4 of MMC1 register 1 (and not register 2, because
page size is 8K) controls whether the WRAM is enabled or not. It must be low
to be enabled. When the WRAM is disabled, reading from and writing to it will
not be possible.
*/


void map001::write(uint32 offset, uint8 value)
{
   int reg = (offset >> 13);

   if (value & 0x80)
   {
      mRegs[0] |= 0x0c;
      mBitCount = 0;
      mLatch = 0;
      return;
   }

   if (mLastReg != reg)
   {
      mLastReg = reg;
      mBitCount = 0;
      mLatch = 0;
   }

   mLatch |= (value & 1) << mBitCount++;

   if (5 != mBitCount)
      return;

   mBitCount = 0;
   mRegs[reg] = mLatch;
   value = mLatch;
   mLatch = 0;

   switch (reg)
   {
   case 0:
      if (0 == (value & 2))
      {
         int mirror = value & 1;
         ((NES *) mpMachine)->getPPU()->mirror(mirror, mirror, mirror, mirror);
      }
      else
      {
         if (value & 1)
            ((NES *) mpMachine)->getPPU()->mirror(0, 0, 1, 1);
         else
            ((NES *) mpMachine)->getPPU()->mirror(0, 1, 0, 1);
      }
      break;

   case 1:
      if (mRegs[0] & 0x10)
         chrbank(4, 0x0000, value);
      else
         chrbank(8, 0x0000, value >> 1);
      break;

   case 2:
      if (mRegs[0] & 0x10)
         chrbank(4, 0x1000, value);
      break;

   case 3:
      /* TODO: hmmmmm... */
      if (mInfo.mNumPRG == 0x20)
      {
         mBankSelect = (mRegs[1] & 0x10) ? 0 : 0x10;
      }
      else if (mInfo.mNumPRG == 0x40)
      {
         if (mRegs[0] & 0x10)
            mBankSelect = (mRegs[1] & 0x10) | ((mRegs[2] & 0x10) << 1);
         else
            mBankSelect = (mRegs[1] & 0x10) << 1;
      }
      else
      {
         mBankSelect = 0;
      }

      if (0 == (mRegs[0] & 0x08))
         prgbank(32, 0x8000, ((mRegs[3] >> 1) + (mBankSelect >> 1)));
      else if (mRegs[0] & 0x04)
         prgbank(16, 0x8000, ((mRegs[3] & 0xf) + mBankSelect));
      else
         prgbank(16, 0xc000, ((mRegs[3] & 0xf) + mBankSelect));

      break;

   default:
      break;
   }
}


void map001::dispatchWrite(void *userdata, uint32 offset, uint8 value)
{
   ((map001 *) userdata)->write(offset, value);
}


static nes6502_memwrite map001_memwrite[] =
{
   { 0x8000, 0xffff, map001::dispatchWrite, NULL },
   { (uint32) -1, (uint32) -1, NULL, NULL }
};


map001::map001(machine *pMachine, mapperInfo *pInfo)
: mmc(pMachine, pInfo)
{
   mpName = "Nintendo MMC1";
   mpMemWrite = map001_memwrite;

   resetRegs();
};


void map001::resetRegs()
{
   for (int i = 0; i < 4; i++)
      mRegs[i] = 0;
   mBitCount = 0;
   mLatch = 0;
   mLastReg = 0;
   mBankSelect = 0;
}


void map001::reset(bool hard_reset)
{
   mmc::reset(hard_reset);

   resetRegs();

   /* TODO: hmmm.... */
   if (mInfo.mNumPRG == 0x20)
      prgbank(16, 0xc000, 0x0f);

   write(0x8000, 0x80);
}


void map001::getState(SnssMapperBlock *pState)
{
   pState->extraData.mapper1.registers[0] = mRegs[0];
   pState->extraData.mapper1.registers[1] = mRegs[1];
   pState->extraData.mapper1.registers[2] = mRegs[2];
   pState->extraData.mapper1.registers[3] = mRegs[3];
   pState->extraData.mapper1.latch = mLatch;
   pState->extraData.mapper1.numberOfBits = mBitCount;
}


void map001::setState(SnssMapperBlock *pState)
{
   mRegs[0] = pState->extraData.mapper1.registers[0];
   mRegs[1] = pState->extraData.mapper1.registers[1];
   mRegs[2] = pState->extraData.mapper1.registers[2];
   mRegs[3] = pState->extraData.mapper1.registers[3];
   mLatch = pState->extraData.mapper1.latch;
   mBitCount = pState->extraData.mapper1.numberOfBits;
}

