//
// shatbox (C) 2002 Matthew Conte (matt@conte.com)
//
//
// nes.cpp
//
// main NES emulation
//
// $Id: $
//

#include <iostream>
#include <string.h>
#include <time.h>
#include <sys/time.h>
#include "assert.h"
#include "nes.h"
#include "ppu.h"
#include "apu.h"
#include "cpu.h"
#include "event.h"
#include "input.h"
#include "mmc.h"
#include "cart.h"
#include "system.h"

NES::NES(const char *pFilename)
: machine(),
  mpROMFilename(pFilename)
{
   mpCart = new cart;
   if (mpCart->load(pFilename))
   {
      delete mpCart;
      std::cerr << "error loading file '" << pFilename << "'." << std::endl;
      exit(-1);
   }

   mpMMC = mmc_create(this, mpCart->getMapperInfo());
   if (NULL == mpMMC)
   {
      delete mpCart;
      std::cerr << "unrecognized mapper hardware." << std::endl;
      exit(-1);
   }

   mpSystem = System_factory();
   if (NULL == mpSystem)
   {
      delete mpMMC;
      delete mpCart;
      std::cerr << "error initializing system" << std::endl;
      exit(-1);
   }

   mpCPU = new cpu(this);
   mpAPU = new apu(this, mpCPU, mpSystem->getSampleRate());
   mpPPU = new ppu(this);
   mpInput = new input(this);

   mpCPU->addTraps(mpAPU);
   mpCPU->addTraps(mpPPU);
   mpCPU->addTraps(mpMMC);
   mpCPU->addTraps(mpInput);
   mpCPU->closeTraps();

   mpEventHandler = new eventHandler(this);

   mFramePeriod = (1000000.0 / 60.0) - (float) mpSystem->timerResolution();

   reset(true);
}


void NES::setColor(int index, int r, int g, int b, int value)
{
   mpSystem->setColor(index, r, g, b, value);
}


NES::~NES()
{
   delete mpSystem;

   delete mpEventHandler;
   delete mpInput;

   delete mpPPU;
   delete mpAPU;
   delete mpCPU;
   delete mpMMC;
   delete mpCart;
}


void  NES::setThrottle(bool throttle)
{
   if (throttle != mThrottle)
   {
      mThrottle = throttle;

      // we're unthrottling, so clear out audio buffer.
      mpSystem->silenceAudio();
   }
}


void  NES::reset(bool hardReset /* = false */)
{
   mpMMC->reset(hardReset);
   mpPPU->reset(hardReset);
   mpAPU->reset(hardReset);
   mpCPU->reset(hardReset);
   mpInput->reset(hardReset);
}


void  NES::triggerInput(int device, int data, bool state)
{
   input::eInputDevice inputDevice;
   input::eInputState inputState = state ? input::INPUT_PRESS : input::INPUT_RELEASE;

   switch (device)
   {
   case 0:
      inputDevice = input::INPUT_PAD0;
      break;
   case 1:
      inputDevice = input::INPUT_PAD1;
      break;
   default:
      ASSERT(0);
      return;
   }

   mpInput->trigger(inputDevice, data, inputState);
}


void  NES::frame()
{
   /*
   // emulate vblank
   int line = mpPPU->currentLine();
   while (line < 262)
   {
      mpMMC->hblank(1);

      // TODO: ugly
      //if (line < 240)
      mpSystem->submitVideo(mpPPU->scanline(), line);

      if (241 == line)
      {
         mpCPU->vbl();

         if (mpPPU->vbl())
            mpCPU->nmi();

         mpMMC->vblank();
      }

      mpCPU->scanline();
      line++;
   }

   // emulate frame
   line = mpPPU->currentLine();
   while (line < 241)
   {
      mpMMC->hblank(0);

      // TODO: ugly
      //if (line < 240)
      mpSystem->submitVideo(mpPPU->scanline(), line);

      mpCPU->scanline();
      line++;
   }
   */

   bool vblank = true;
   bool frameComplete = false;
   while (!frameComplete)
   {
      mpMMC->hblank(vblank);

      int scanline = mpPPU->currentLine();
      uint8 *pBuf = mpPPU->scanline();

      switch (scanline)
      {
      case 240:
         frameComplete = true;
         break;

      case 241:
         mpCPU->vbl();
         if (mpPPU->vbl())
            mpCPU->nmi();
         mpMMC->vblank();
         break;

      case 261:
         vblank = false;
         break;

      default:
         if (scanline < 240)
            mpSystem->submitVideo(pBuf, scanline);
         break;
      }

      mpCPU->scanline();
   }

   // final sound frame update
   mpAPU->update();
}


void  NES::run()
{
   int frames = 0;

   clock_t elapsed, start = clock();

   mpSystem->mark();

   for (;;)
   {
      if (mpSystem->pollEvents(mpEventHandler))
         break;

      if (mpEventHandler->process())
         break;

      mpSystem->lockVideo();
      frame();
      mpSystem->unlockVideo();

      unsigned char *pStream;
      int samples = mpAPU->flushSamples(&pStream);

      if (mThrottle)
      {
         mpSystem->submitAudio(pStream, samples);

         int delay = (int) mFramePeriod - mpSystem->elapsed();
         mpSystem->usleep(delay);
         mpSystem->mark();
      }
      else
      {
         mpSystem->silenceAudio();
      }

      frames++;
   }

   elapsed = clock() - start;
   if (elapsed)
   {
      float fps = frames / (elapsed / (float) CLOCKS_PER_SEC);
      std::cout << "average fps: " << fps << std::endl;
   }
}


void NES::saveState()
{
}


void NES::loadState()
{
}
